/*
Module : CEFWrappers.cpp
Purpose: Defines the implementation for MFC wrapper classes for the Chromium Embedded Framework aka CEF
Created: PJN / 09-02-2020
History: PJN / 09-02-2020 1. Initial Creation
         PJN / 04-03-2020 1. Updated the demo application to only create the CCefAppHandler instance prior to calling
                          CefInitialize in the CCEFDemoApp::InitializeCef method in the demo app. This is a small 
                          optimization to make the demo application code more consistent with the cefsimple application.
                          2. Updated the CCefAppHandler class to support the CefBrowserProcessHandler interface.
                          3. Updated the CCEFView::SetupClientHandler method to not set the web_security settings value
                          to STATE_DISABLED. This allows the demo program to better render some web pages.
                          4. Updated the sample application to include x86 build configurations in addition to the 
                          existing x64 build configurations.
                          5. Updated the demo application to load CEF from a subdirectory called "CEF" below where the 
                          main exe is located. To see how this is done, please see the code in CCEFDemoApp::InitInstance.
                          6. Made CCEFView::SetupClientHandler method virtual
         PJN / 20-08-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.
                          3. Added SAL annotations throughout the codebase.
                          4. Updated the compiler and linker settings of the demo app to more modern defaults.
         PJN / 26-12-2023 1. Updated copyright details.
                          2. Updated code to pull in CEF v120.1.10

Copyright (c) 2020 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "CEFWrappers.h"


//////////////////// Statics / Macros /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CCEFView, CView)

#pragma warning(suppress: 26440 26433)
BEGIN_MESSAGE_MAP(CCEFView, CView) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_SIZE()
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
CCEFView::CCEFView() : m_sStartURL{_T("http://tests/intro.html")}
{
}

CCEFView::~CCEFView()
{
#pragma warning(suppress: 26447)
  if (m_ClientHandler != nullptr)
    m_ClientHandler->DetachDelegate();
}

#pragma warning(suppress: 26440)
void CCEFView::OnDraw(CDC* /*pDC*/)
{
  //Nothing to do
}

#pragma warning(suppress: 26434)
void CCEFView::OnSize(UINT nType, int cx, int cy)
{
  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);

  //Resize the CEF window to be the size of the client area
  if (m_Browser != nullptr)
  {
    ASSERT(m_Browser->GetHost() != nullptr);
#pragma warning(suppress: 26486)
    auto hWnd{m_Browser->GetHost()->GetWindowHandle()};
    CRect rect;
    GetClientRect(&rect);
#pragma warning(suppress: 26477)
    ::SetWindowPos(hWnd, HWND_TOP, rect.left, rect.top, rect.Width(), rect.Height(), SWP_NOZORDER);
  }
}

void CCEFView::SetupClientHandler(_In_opt_ CCEFClientHandler* pClientHandler)
{
  //Create the CEF browser
  CRect rect;
  GetClientRect(&rect);
  const CefRect cefRect{rect.left, rect.top, rect.Width(), rect.Height()};
  CefWindowInfo cefInfo;
  cefInfo.SetAsChild(GetSafeHwnd(), cefRect);
  CefBrowserSettings cefSettings;
  ASSERT(m_ClientHandler == nullptr);
#pragma warning(suppress: 26409)
  m_ClientHandler = pClientHandler;
  m_ClientHandler->CreateBrowser(cefInfo, cefSettings, CefString{m_sStartURL});
}

void CCEFView::OnAfterCreated(_Inout_ CefRefPtr<CefBrowser> browser)
{
  //Cache the browser object
  m_Browser = browser;
}

#pragma warning(suppress: 26415 26418)
void CCEFView::OnBeforeClose(_Inout_ CefRefPtr<CefBrowser> browser)
{
  if ((m_Browser != nullptr) && (m_Browser->GetIdentifier() == browser->GetIdentifier()))
  {
    m_Browser = nullptr;
    m_ClientHandler->DetachDelegate();
  }
}

BOOL CCEFView::PreTranslateMessage(MSG* pMsg)
{
  //Validate our parameters
  ASSERT(pMsg != nullptr);

  //Support for Refresh/Reload via the F5 key
  if ((m_Browser != nullptr) && (pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_F5))
    m_Browser->Reload();

  return __super::PreTranslateMessage(pMsg);
}

void CCEFView::Navigate(_In_ const CString& sURL)
{
  //Validate our parameters
  ASSERT(m_Browser != nullptr);

  auto pFrame{m_Browser->GetMainFrame()};
  if (pFrame != nullptr)
#pragma warning(suppress: 26486)
    pFrame->LoadURL(CefString{sURL});
}

void CCEFView::NavigateBack()
{
  //Validate our parameters
  ASSERT(m_Browser != nullptr);

  m_Browser->GoBack();
}

void CCEFView::NavigateForward()
{
  //Validate our parameters
  ASSERT(m_Browser != nullptr);

  m_Browser->GoForward();
}

bool CCEFView::CanNavigateBack()
{
  return (m_Browser != nullptr) && m_Browser->CanGoBack();
}

bool CCEFView::CanNavigateForward()
{
  return (m_Browser != nullptr) && m_Browser->CanGoForward();
}

void CCEFView::CloseBrowser()
{
  if (m_Browser != nullptr)
  {
    ASSERT(m_Browser->GetHost() != nullptr);
#pragma warning(suppress: 26486)
    ::DestroyWindow(m_Browser->GetHost()->GetWindowHandle());
  }
}


#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CCEFDoc, CDocument)

#pragma warning(suppress: 26455)
CCEFDoc::CCEFDoc()
{
  ASSERT_VALID(this);
}

CCEFView* CCEFDoc::GetView() const
{
  //find the first view - if there are no views
  //we must return nullptr
  POSITION pos{GetFirstViewPosition()};
  if (pos == nullptr)
    return nullptr;

  //find the first view that is a CCEFView
  while (pos != nullptr)
  {
#pragma warning(suppress: 26429)
    CView* pView{GetNextView(pos)};
    ASSERT(pView != nullptr);
    if (pView->IsKindOf(RUNTIME_CLASS(CCEFView)))
#pragma warning(suppress: 26466)
      return static_cast<CCEFView*>(pView);
  }

  //can't find one, then return nullptr
  return nullptr;
}
