/*
Module : CEFWrappers.h
Purpose: Defines the interface for MFC wrapper classes for the Chromium Embedded Framework aka CEF
Created: PJN / 09-02-2020

Copyright (c) 2020 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error CEFWrappers requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __CEFWRAPPERS_H__
#define __CEFWRAPPERS_H__

#ifndef CEFWRAPPERS_EXT_CLASS
#define CEFWRAPPERS_EXT_CLASS
#endif //#ifndef CEFWRAPPERS_EXT_CLASS


//////////////////// Classes //////////////////////////////////////////////////

class CCEFDelegate
{
public:
//Constructors / Destructors
  CCEFDelegate() = default;
  CCEFDelegate(const CCEFDelegate&) = delete;
  CCEFDelegate(CCEFDelegate&&) = delete;
  virtual ~CCEFDelegate() = default;

//Methods
  CCEFDelegate& operator=(const CCEFDelegate&) = delete;
  CCEFDelegate& operator=(CCEFDelegate&&) = delete;

#pragma warning(suppress: 26440)
  virtual void OnSetAddress(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _Inout_ CefRefPtr<CefFrame> /*frame*/, _In_ const std::string& /*url*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnSetTitle(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ const std::string& /*title*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnFaviconURLChange(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ const std::vector<CefString>& /*icon_urls*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnFullscreenModeChange(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ bool /*fullscreen*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual bool OnTooltip(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _Inout_ CefString& /*text*/)
  {
    return false;
  }

#pragma warning(suppress: 26440)
  virtual void OnStatusMessage(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ const CefString& /*value*/)
  {
  }

#pragma warning(suppress: 26440 26812)
  virtual bool OnConsoleMessage(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ cef_log_severity_t /*level*/, _In_ const CefString& /*message*/, _In_ const CefString& /*source*/, _In_ int /*line*/)
  {
    return false;
  }

#pragma warning(suppress: 26440)
  virtual bool OnAutoResize(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ const CefSize& /*new_size*/)
  {
    return false;
  }

#pragma warning(suppress: 26440)
  virtual void OnLoadingProgressChange(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ double /*progress*/)
  {
  }

#pragma warning(suppress: 26440 26812)
  virtual bool OnBeforePopup(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _Inout_ CefRefPtr<CefFrame> /*frame*/, _In_ const CefString& /*target_url*/,
                             _In_ const CefString& /*target_frame_name*/, _In_ CefLifeSpanHandler::WindowOpenDisposition /*target_disposition*/, _In_ bool /*user_gesture*/,
                             _In_ const CefPopupFeatures& /*popupFeatures*/, _Inout_ CefWindowInfo& /*windowInfo*/, _Inout_ CefRefPtr<CefClient>& /*client*/,
                             _Inout_ CefBrowserSettings& /*settings*/, _Inout_ CefRefPtr<CefDictionaryValue>& /*extra_info*/, _Inout_ bool* /*no_javascript_access*/)
  {
    return false;
  }

#pragma warning(suppress: 26440)
  virtual void OnAfterCreated(_Inout_ CefRefPtr<CefBrowser> /*browser*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual bool DoClose(_Inout_ CefRefPtr<CefBrowser> /*browser*/)
  {
    return false;
  }

#pragma warning(suppress: 26440)
  virtual void OnBeforeClose(_Inout_ CefRefPtr<CefBrowser> /*browser*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnLoadingStateChange(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _In_ bool /*isLoading*/, _In_ bool /*canGoBack*/, _In_ bool /*canGoForward*/)
  {
  }

#pragma warning(suppress: 26440 26812)
  virtual void OnLoadStart(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _Inout_ CefRefPtr<CefFrame> /*frame*/, _In_ CefLoadHandler::TransitionType /*transition_type*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnLoadEnd(CefRefPtr<CefBrowser> /*browser*/, CefRefPtr<CefFrame> /*frame*/, int /*httpStatusCode*/)
  {
  }

#pragma warning(suppress: 26415 26418 26812)
  virtual void OnLoadError(_Inout_ CefRefPtr<CefBrowser> /*browser*/, _Inout_ CefRefPtr<CefFrame> frame, _In_ CefLoadHandler::ErrorCode errorCode,
                           _In_ const CefString& errorText, _In_ const CefString& failedUrl)
  {
    //Don't display an error for downloaded files
    if (errorCode == ERR_ABORTED)
      return;

    //Display a load error message
    std::stringstream ss;
    ss << "<html><body>"
          "<p>Failed to download URL " << std::string{failedUrl} << 
          " with error " << std::string{errorText} << " (" << errorCode << 
          ").</p></body></html>";
    frame->LoadURL(GetDataURI(ss.str(), "text/html"));
  }

  std::string GetDataURI(_In_ const std::string& data, _In_ const std::string& mime_type)
  {
    return "data:" + mime_type + ";base64," + CefURIEncode(CefBase64Encode(data.data(), data.size()), false).ToString();
  }
};


class CEFWRAPPERS_EXT_CLASS CCEFClientHandler : public CefClient,
                                                public CefDisplayHandler,
                                                public CefLifeSpanHandler,
                                                public CefLoadHandler,
                                                public CefRequestHandler,
                                                public CefResourceRequestHandler
{
public:
//Constructors / Destructors
  CCEFClientHandler(_In_opt_ CCEFDelegate* pDelegate) : m_pDelegate{pDelegate}
  {
  }
  CCEFClientHandler() = delete;
  CCEFClientHandler(const CCEFClientHandler&) = delete;
  CCEFClientHandler(CCEFClientHandler&&) = delete;
  ~CCEFClientHandler() = default; //NOLINT(modernize-use-override)

//Methods
  CCEFClientHandler& operator=(const CCEFClientHandler&) = delete;
  CCEFClientHandler& operator=(CCEFClientHandler&&) = delete;

  void CreateBrowser(_In_ const CefWindowInfo& info, _In_ const CefBrowserSettings& settings, _In_ const CefString& sURL,
                     _Inout_ CefRefPtr<CefDictionaryValue> extra_info = nullptr, _Inout_ CefRefPtr<CefRequestContext> request_context = nullptr)
  {
    CefBrowserHost::CreateBrowser(info, this, sURL, settings, extra_info, request_context);
  }

  void DetachDelegate() noexcept
  {
    m_pDelegate = nullptr;
  }

//CefClient methods
  CefRefPtr<CefDisplayHandler> GetDisplayHandler() override
  {
    return this;
  }

  CefRefPtr<CefLifeSpanHandler> GetLifeSpanHandler() override
  {
    return this;
  }

  CefRefPtr<CefLoadHandler> GetLoadHandler() override
  {
    return this;
  }

  //Note currently CCEFClientHandler does not support GetContextMenuHandler, GetDialogHandler,
  //GetDownloadHandler, GetDragHandler, GetFindHandler, GetFocusHandler, GetJSDialogHandler,
  //GetKeyboardHandler, GetRenderHandler. A future iteration of the code 
  //could support these tear off interfaces in CCEFClientHandler if you think it would be useful.

  //CefDisplayHandler methods
#pragma warning(suppress: 26415 26418)
  void OnAddressChange(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, const CefString& url) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (frame->IsMain() && (m_pDelegate != nullptr))
      m_pDelegate->OnSetAddress(browser, frame, url);
  }

#pragma warning(suppress: 26418)
  void OnTitleChange(CefRefPtr<CefBrowser> browser, const CefString& title) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnSetTitle(browser, title);
  }

  void OnFaviconURLChange(CefRefPtr<CefBrowser> browser, const std::vector<CefString>& icon_urls) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnFaviconURLChange(browser, icon_urls);
  }

  void OnFullscreenModeChange(CefRefPtr<CefBrowser> browser, bool fullscreen) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnFullscreenModeChange(browser, fullscreen);
  }

  bool OnTooltip(CefRefPtr<CefBrowser> browser, CefString& text) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      return m_pDelegate->OnTooltip(browser, text);
    else
      return __super::OnTooltip(browser, text);
  }

  void OnStatusMessage(CefRefPtr<CefBrowser> browser, const CefString& value) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnStatusMessage(browser, value);
  }

  bool OnConsoleMessage(CefRefPtr<CefBrowser> browser, cef_log_severity_t level, const CefString& message, const CefString& source, int line) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      return m_pDelegate->OnConsoleMessage(browser, level, message, source, line);
    else
      return __super::OnConsoleMessage(browser, level, message, source, line);
  }

  bool OnAutoResize(CefRefPtr<CefBrowser> browser, const CefSize& new_size) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      return m_pDelegate->OnAutoResize(browser, new_size);
    else
      return __super::OnAutoResize(browser, new_size);
  }

  void OnLoadingProgressChange(CefRefPtr<CefBrowser> browser, double progress) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnLoadingProgressChange(browser, progress);
  }

//CefLifespanHandler methods
  bool OnBeforePopup(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, const CefString& target_url,
                     const CefString& target_frame_name, CefLifeSpanHandler::WindowOpenDisposition target_disposition, bool user_gesture,
                     const CefPopupFeatures& popupFeatures, CefWindowInfo& windowInfo, CefRefPtr<CefClient>& client,
                     CefBrowserSettings& settings, CefRefPtr<CefDictionaryValue>& extra_info, bool* no_javascript_access) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      return m_pDelegate->OnBeforePopup(browser, frame, target_url,
                                        target_frame_name, target_disposition, user_gesture,
                                        popupFeatures, windowInfo, client,
                                        settings, extra_info, no_javascript_access);
    else
      return __super::OnBeforePopup(browser, frame, target_url,
                                    target_frame_name, target_disposition, user_gesture,
                                    popupFeatures, windowInfo, client,
                                    settings, extra_info, no_javascript_access);
  }

  void OnAfterCreated(CefRefPtr<CefBrowser> browser)  override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnAfterCreated(browser);
  }

  bool DoClose(CefRefPtr<CefBrowser> browser) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      return m_pDelegate->DoClose(browser);
    else
      return __super::DoClose(browser);
  }

  void OnBeforeClose(CefRefPtr<CefBrowser> browser) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnBeforeClose(browser);
  }

//CefLoadHandler methods
  void OnLoadingStateChange(CefRefPtr<CefBrowser> browser, bool isLoading, bool canGoBack, bool canGoForward) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnLoadingStateChange(browser, isLoading, canGoBack, canGoForward);
  }

  void OnLoadStart(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, TransitionType transition_type) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnLoadStart(browser, frame, transition_type);
  }

  void OnLoadEnd(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, int httpStatusCode) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnLoadEnd(browser, frame, httpStatusCode);
  }

  void OnLoadError(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, ErrorCode errorCode,
                   const CefString& errorText, const CefString& failedUrl) override
  {
    CEF_REQUIRE_UI_THREAD();

    if (m_pDelegate != nullptr)
      m_pDelegate->OnLoadError(browser, frame, errorCode, errorText, failedUrl);
  }

protected:
//Member variables
  CCEFDelegate* m_pDelegate;

  IMPLEMENT_REFCOUNTING(CCEFClientHandler);
};


class CEFWRAPPERS_EXT_CLASS CCEFView : public CView, public CCEFDelegate
{
public:
//Constructors / Destructors
  CCEFView();
  CCEFView(const CCEFView&) = delete;
  CCEFView(CCEFView&&) = delete;
  ~CCEFView(); //NOLINT(modernize-use-override)

//Methods
  CCEFView& operator=(const CCEFView&) = delete;
  CCEFView& operator=(CCEFView&&) = delete;
  void Navigate(_In_ const CString& sURL);
  void NavigateBack();
  void NavigateForward();
  bool CanNavigateBack();
  bool CanNavigateForward();
  void CloseBrowser();
  void SetStartURL(_In_ const CString& sStartURL) { m_sStartURL = sStartURL; };
  [[nodiscard]] CString GetStartURL() const { return m_sStartURL; };

protected:
//Methods
  void OnAfterCreated(_Inout_ CefRefPtr<CefBrowser> browser) override;
  void OnBeforeClose(_Inout_ CefRefPtr<CefBrowser> browser) override;
  void OnDraw(CDC* pDC) override;
  BOOL PreTranslateMessage(MSG* pMsg) override;
  virtual void SetupClientHandler(_In_opt_ CCEFClientHandler* pClientHandler);

//Message handlers
  afx_msg void OnSize(UINT nType, int cx, int cy);

//Member variables
  CString m_sStartURL;
  CefRefPtr<CCEFClientHandler> m_ClientHandler;
  CefRefPtr<CefBrowser> m_Browser;

  DECLARE_MESSAGE_MAP()
  DECLARE_DYNCREATE(CCEFView)

  friend class CCEFDoc;
};


class CEFWRAPPERS_EXT_CLASS CCEFDoc : public CDocument
{
protected: //create from serialization only
  CCEFDoc();

public:
//Methods
  [[nodiscard]] virtual CCEFView* GetView() const;

protected:
  DECLARE_DYNAMIC(CCEFDoc)
};


#endif //#ifndef __CEFWRAPPERS_H__
